package org.tbasket.handler

import org.apache.logging.log4j.{LogManager, Logger}
import org.tbasket.auth.Authenticator
import org.tbasket.data.User
import org.tbasket.error.*
import org.tbasket.handler.HandlerUtils.{errorBody, parseAttribute, parseRequestForm, search}
import zio.ZIO
import zio.http.model.{Cookie, Headers, Status}
import zio.http.{Body, Request, Response, model}
import zio.json.*
import zio.json.ast.{Json, JsonCursor}

object RegisterPageHandler extends PageHandler {
  
  implicit private final val Log: Logger = LogManager.getLogger("/register")
  
  private def tryPost(request: Request) =
    for
      params <- parseRequestForm(request.body)
      
      name <- search(params)("name")
      forename <- search(params)("forename")
      mail <- search(params)("email")
      password <- search(params)("password")
      
      user <- ZIO.serviceWithZIO[Authenticator](_.registerUser(name, forename, mail, password))
      jwt <- ZIO.serviceWithZIO[Authenticator](_.requestNewJwt(user))
    yield Response(
      status = Status.Found,
      headers = Headers.location("/") ++ //register successful, go back to main page
        Headers.setCookie(Cookie("JWT", jwt))
    )
  
  def post(request: Request) = tryPost(request)
    .catchSome {
      case InvalidEmail(msg)          => ZIO.attempt(Response(
        status = Status.ExpectationFailed,
        body = errorBody("invalid email", msg)
      ))
      case InvalidPassword(msg)       => ZIO.attempt(Response(
        status = Status.ExpectationFailed,
        body = errorBody("invalid password", msg)
      ))
      case UserAlreadyRegistered(msg) =>
        ZIO.attempt(Response(
          status = Status.NotAcceptable,
          body = errorBody("already registered", msg),
          headers = Headers.location("/login") //the account already exists so we move the user to login page
        ))
    }
  
}
